# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Ryan Robison
# Date: 2013Aug

import gpi


class ExternalNode(gpi.NodeAPI):
    """Combine two data sets along any dimension.
    e.g. put two images side-by-side
    
    Several restrictions are in place requiring arrays to be of the same size;
      this module should be updated to be more flexible in the future
    """

    def initUI(self):
        # Widgets
        self.addWidget('TextBox', 'Info', val='Ready')
        self.addWidget('Slider', 'Combine Dimension',min=-1,max=0,val=0)

        # IO Ports
        self.addInPort('indata1', 'NPYarray', obligation=gpi.REQUIRED)
        self.addInPort('indata2', 'NPYarray', obligation=gpi.REQUIRED)

        self.addOutPort('outdata', 'NPYarray')

    def validate(self):
        
        indata1  = self.getData('indata1')
        indata2  = self.getData('indata2')
        comb_dim = self.getVal('Combine Dimension')
        self.setAttr('Combine Dimension', max = indata1.ndim, val=comb_dim)

        info = "input1 : " +str(indata1.shape) + "\n" + "input2 : " +str(indata2.shape)
        self.setAttr('Info', val=info)

        return 0

    def compute(self):
        '''This is where the main algorithm should be implemented.
        '''
        import numpy as np

        indata1  = self.getData('indata1')
        indata2  = self.getData('indata2')
        comb_dim = self.getVal('Combine Dimension')
        
        if comb_dim == indata1.ndim:
            temp1 = np.expand_dims(indata1, comb_dim)
            temp2 = np.expand_dims(indata2, comb_dim)

            try:
                outdata = np.append(temp1, temp2, axis = comb_dim)
            except ValueError as dim_err:
                self.log.warn('Combine error: '+str(dim_err))
            else:
                info = "input1 : " +str(indata1.shape) + "\ninput2 : " +str(indata2.shape) + "\noutput : " +str(outdata.shape)
                self.setAttr('Info', val=info)
                self.setData('outdata',outdata)
        elif comb_dim == -1 :
            temp1 = np.expand_dims(indata1, 0)
            temp2 = np.expand_dims(indata2, 0)

            try:
                outdata = np.append(temp1, temp2, axis = 0)
            except ValueError as dim_err:
                self.log.warn('Combine error: '+str(dim_err))
            else:
                info = "input1 : " +str(indata1.shape) + "\ninput2 : " +str(indata2.shape) + "\noutput : " +str(outdata.shape)
                self.setAttr('Info', val=info)
                self.setData('outdata',outdata)
        else:
            try:
                outdata = np.append(indata1, indata2, axis = comb_dim)
            except ValueError as dim_err:
                self.log.warn('Combine error: '+str(dim_err))
            else:
                info = "input1 : " +str(indata1.shape) + "\ninput2 : " +str(indata2.shape) + "\noutput : " +str(outdata.shape)
                self.setAttr('Info', val=info)
                self.setData('outdata',outdata)

        
        return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart
# Date: 2013aug16

import gpi


class ExternalNode(gpi.NodeAPI):
    """A node for sharing data, code, and the joy of GPI with others.
        -keep others up to date with step-by-step tweets of your exec-flow.
        -ping your blogosphere with the latest data analysis.
    """

    def initUI(self):
        # Widgets
        self.addWidget('WebBox', 'Facebook', val='https://www.facebook.com')
        self.addWidget('WebBox', 'Twitter', val='https://twitter.com')

    def execType(self):
        return gpi.GPI_THREAD



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart
# Date: 2012sep02

import gpi
from gpi import QtWidgets

# WIDGET


class ReduceSliders(gpi.GenericWidgetGroup):
    """Combines the BasicCWFCSliders with ExclusivePushButtons
    for a unique widget element useful for reduce dimensions.
    """
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        super().__init__(title, parent)
        self.sl = gpi.BasicCWFCSliders()
        self.sl.valueChanged.connect(self.valueChanged)
        # at least one button
        self.button_names = ['C/W', 'B/E', 'Slice', 'Pass']
        self.buttons = []
        cnt = 0
        wdgLayout = QtWidgets.QGridLayout()
        for name in self.button_names:
            newbutton = QtWidgets.QPushButton(name)
            newbutton.setCheckable(True)
            newbutton.setAutoExclusive(True)
            self.buttons.append(newbutton)
            newbutton.clicked.connect(self.findValue)
            newbutton.clicked.connect(self.valueChanged)
            wdgLayout.addWidget(newbutton, 0, cnt, 1, 1)
            cnt += 1
        # layout
        wdgLayout.addWidget(self.sl, 1, 0, 1, 4)
        wdgLayout.setVerticalSpacing(0)
        wdgLayout.setSpacing(0)
        self.setLayout(wdgLayout)
        # default
        self.set_min(1)
        self._selection = 3  # pass is default
        self.buttons[self._selection].setChecked(True)
        self.sl.set_allvisible(False)

    # setters
    def set_val(self, val):
        """A python-dict containing keys: center, width, floor, ceiling,
        and selection.
        """
        if 'center' in val:
            self.sl.set_center(val['center'])
        if 'width' in val:
            self.sl.set_width(val['width'])
        if 'floor' in val:
            self.sl.set_floor(val['floor'])
        if 'ceiling' in val:
            self.sl.set_ceiling(val['ceiling'])
        if 'selection' in val:
            self._selection = val['selection']
        self.buttons[self._selection].blockSignals(True)
        self.buttons[self._selection].setChecked(True)
        self.buttons[self._selection].blockSignals(False)
        self.findValue(None)

    def set_min(self, val):
        """A min value for center, width, floor and ceiling (int)."""
        self.sl.set_min(val)

    def set_max(self, val):
        """A max value for center, width, floor and ceiling (int)."""
        self.sl.set_max(val)

    def set_quietmax(self, val):
        if val is not None:
            self.blockSignals(True)
            self.set_max(val)
            self.blockSignals(False)

    # getters
    def get_val(self):
        val = {}
        val['selection'] = self._selection
        val['center'] = self.sl.get_center()
        val['width'] = self.sl.get_width()
        val['floor'] = self.sl.get_floor()
        val['ceiling'] = self.sl.get_ceiling()
        return val

    def get_min(self):
        return self.sl.get_min()

    def get_max(self):
        return self.sl.get_max()

    def get_quietmax(self):
        # don't return a value so that network deserialize passes 'None' to
        # the setter
        pass

    # support

    def setCropBounds(self):
        self.sl.set_min_width(1)

    def setSliceBounds(self):
        self.sl.set_min_width(0)
        self.sl.set_width(1)

    def setPassBounds(self):
        self.sl.set_min_width(1)
        self.sl.set_width(self.get_max())
        # JGP For Pass, reset cwfc
        self.sl.set_center((self.get_max()+1)//2)
        self.sl.set_width(self.get_max())
        self.sl.set_floor(1)
        self.sl.set_ceiling(self.get_max())

    def findValue(self, value):
        cnt = 0
        for button in self.buttons:
            if button.isChecked():
                self._selection = cnt
            cnt += 1
        # hide appropriate sliders
        if self._selection == 0:  # C/W
            self.sl.set_cwvisible(True)
            self.setCropBounds()
        elif self._selection == 1:  # B/E
            self.sl.set_fcvisible(True)
            self.setCropBounds()
        elif self._selection == 2:  # slice
            self.sl.set_slicevisible(True)
            self.setSliceBounds()
        else:  # pass
            self.sl.set_allvisible(False)
            self.setPassBounds()


class ExternalNode(gpi.NodeAPI):
    """A module for slicing, cropping and masking n-D numpy arrays.

    INPUT - input array

    OUTPUTS: note the "out" port is the data you typically want, while the "mask" port shows where that data came from
    out - data after cropping or slicing - size may be different than input array
    mask - copy of input array (same size), but data replaced with zeros wherever data were cropped/sliced

    WIDGETS:
    I/O info: - shows size of input, output arrays
    Dimension[i]
      C/W - sliders select the center and width of cropping range along the ith dimension
      B/E - sliders select the beginning and end of cropping range along the ith dimension
      Slice - slider selects the index to slice along the ith dimension
      Pass - ith dimension is passed (not affected)
    Mask - generate data for "mask" output - good to leave off for large data sets if not needed
    Compute - generate sliced/cropped data

    The Reduce Widget takes a dict with the following keys:
        d = {}
        d['selection'] # integer 0-3 for 'C/W', 'B/E', 'Slice' and 'Pass'

        d['center']    # integer 1-N only for 'C/W'
        d['width']     # integer 1-N only for 'C/W'

        d['floor']     # integer 1-N only for 'B/E'
        d['ceiling']   # integer 1-N only for 'B/E'
    Not all keys have to be present.
    """
    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS

    def initUI(self):

        # Widgets
        self.addWidget('TextBox', 'I/O Info:')

        self.dim_base_name = 'Dimension['
        self.ndim = 10
        for i in range(self.ndim):
            self.addWidget('ReduceSliders', self.dim_base_name+str(-i-1)+']')
        self.addWidget('PushButton', 'Mask', toggle=True)
        self.addWidget('PushButton', 'Compute', toggle=True, val=True)

        # IO Ports
        self.addInPort('in', 'NPYarray', obligation=gpi.REQUIRED)
        self.addOutPort('out', 'NPYarray')
        self.addOutPort('mask', 'NPYarray')

    def validate(self):
        '''update the widget bounds based on the input data
        '''

        # only update bounds if the 'in' port changed.
        if 'in' in self.portEvents():

            data = self.getData('in')
            dilen = len(data.shape)

            # visibility and bounds
            for i in range(self.ndim):
                if i < dilen:
                    self.setAttr(self.dim_base_name+str(-i-1)+']', visible=True)
                    self.setAttr(self.dim_base_name+str(-i-1)+']', quietmax=data.shape[-i-1])

                    # JGP for Pass, always max out floor and ceiling always
                    w = self.getVal(self.dim_base_name+str(-i-1)+']')
                    if w['selection'] == 3:
                      wmax = data.shape[i]
                      w['center'] = (wmax+1)//2
                      w['width'] = wmax
                      w['floor'] = 1
                      w['ceiling'] = wmax
                      self.setAttr(self.dim_base_name+str(-i-1)+']', quietval=w)
                else:
                    self.setAttr(self.dim_base_name+str(-i-1)+']',
                            visible=False)

        return(0)

    def compute(self):

        if self.getVal('Compute'):

            data = self.getData('in')
            dilen = len(data.shape)

            # setup slicer
            xi = []
            for i in range(dilen-1, -1, -1):
                w = self.getVal(self.dim_base_name+str(-i-1)+']')
                if w['selection'] == 3:  # pass
                    xi.append(slice(None))
                elif w['selection'] == 2:  # slice
                    xi.append(w['center']-1)
                else:  # C/W & B/E
                    xi.append(slice(w['floor']-1, w['ceiling']))

            # apply indices to the data
            out = data[tuple(xi)]

            # update UI info
            self.setAttr('I/O Info:', val="input: "+str(data.shape)
                    +"\noutput: "+str(out.shape))

            self.setData('out', out)

            # do mask operations
            if self.getVal('Mask'):
                mask = data.copy()
                mask[xi] = 0
                self.setData('mask', mask)

        return(0)



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart
# Date: 2013mar03

import os
import gpi


class ExternalNode(gpi.NodeAPI):
    """Warn the user when this node has received data.  Either thru terminal
    bell or OSX voice synth.  The voice settings can be changed in the System Preferences.
    """

    def initUI(self):

        # test for the existence of say
        try:
            if not os.system('/usr/bin/say \"\"'):
                self._say = True
            else:
                self._say = False
        except:
            self._say = False

        # personalize it
        try:
            name = str(os.getlogin()).capitalize()+', y'
        except:
            name = 'Y'

        # Widgets
        self.addWidget('PushButton', 'Voice', toggle=True, val=False, visible=self._say)
        self.addWidget('StringBox', 'Text to Voice', val=name+"our G.P.I. network has reached the Alert node.", visible=self._say)
        self.addWidget('PushButton', 'Terminal Bell', toggle=True, val=False)
        self.addWidget('SpinBox', 'Iterations', val=1, min=1)

        # IO Ports
        self.addInPort('any', 'PASS', obligation=gpi.OPTIONAL)


    def compute(self):
        '''This is where the main algorithm should be implemented.
        '''

        if self.getVal('Voice'):
            for i in range(self.getVal('Iterations')):
                try:
                    os.system('/usr/bin/say \"'+self.getVal('Text to Voice')+'\"')
                except:
                    log.error('Failing to execute \'say\' command.')

        if self.getVal('Terminal Bell'):
            for i in range(self.getVal('Iterations')):
                print('\a')

        return 0



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.

# Author: Jim Pipe / Nick Zwart
# Date: 2013 Sep 01

import gpi
from gpi import QtGui, QtWidgets
import numpy as np

# WIDGET
class WindowLevel(gpi.GenericWidgetGroup):
    """Provides an interface to the BasicCWFCSliders."""
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        super().__init__(title, parent)
        self.sl = gpi.BasicCWFCSliders()
        self.sl.valueChanged.connect(self.valueChanged)
        self.pb = gpi.BasicPushButton()
        self.pb.set_button_title('reset')
        # layout
        wdgLayout = QtWidgets.QVBoxLayout()
        wdgLayout.addWidget(self.sl)
        wdgLayout.addWidget(self.pb)
        self.setLayout(wdgLayout)
        # default
        self.set_min(0)
        self.set_max(100)
        self.sl.set_allvisible(True)
        self.reset_sliders()
        self.pb.valueChanged.connect(self.reset_sliders)

    # setters
    def set_val(self, val):
        """Set multiple values with a python-dict with keys:
        level, window, floor and ceiling. -Requires integer
        values for each key.
        """
        self.sl.set_center(val['level'])
        self.sl.set_width(val['window'])
        self.sl.set_floor(val['floor'])
        self.sl.set_ceiling(val['ceiling'])

    def set_min(self, val):
        """Set min for level, window, floor and ceiling (int)."""
        self.sl.set_min(val)

    def set_max(self, val):
        """Set max for level, window, floor and ceiling (int)."""
        self.sl.set_max(val)

    # getters
    def get_val(self):
        val = {}
        val['level'] = self.sl.get_center()
        val['window'] = self.sl.get_width()
        val['floor'] = self.sl.get_floor()
        val['ceiling'] = self.sl.get_ceiling()
        return val

    def get_min(self):
        return self.sl.get_min()

    def get_max(self):
        return self.sl.get_max()

    def reset_sliders(self):
        val = {}
        val['window'] = 100
        val['level'] = 50
        val['floor'] = 0
        val['ceiling'] = 100
        self.set_val(val)


class ExternalNode(gpi.NodeAPI):
    """2D image viewer for real or complex NPYarrays.

    INPUT:
    2D data, real or complex
    3D uint8 ARGB data (e.g. output of another ImageDisplay node)

    OUTPUT:
    3D data of displayed image, last dimension has length 4 for ARGB byte (uint8) data

    WIDGETS:
    Complex Display - If data are complex, allows you to show Real, Imaginary, Magnitude, Phase, or "Complex" data.
        If C (Complex) is chosen, then pixel brightness reflects value magnitude, while pixel color reflects value phase.
        If input data are real-valued, this widget is hidden

    Color Map - Chooses from a number of colormaps for real-valued data.  Not available if Scalar Display is "Sign"

    Edge Pixels - only visible for complex input data with Complex Display set to "C"
        Setting this to N creates an N-pixel color ring around the image border illustrating the phase-to-color mapping

    Black Pixels - only visible for complex input data with Complex Display set to "C"
        Setting this to N creates an N-pixel black ring around the image border (but inside the Edge Pixels ring) to
        separate the Edge pixel ring from the actual data image

    Viewport - displays the image
      Double clicking on Viewport area brings up a scaling widget to change the image size, and change graphic overlay

    L W F C - (hidden by default - double click on widget area to show sliders)
              Adjust value-to-pixel brightness mapping using Level/Window or Floor/Ceiling

    Scalar Display - visible for real data, or complex data with "Complex Display" set to R, I, M, or P
      Pass uses the real data in the data-to-pixel mapping
      Mag uses the magnitude data in the data-to-pixel mapping (i.e. affects negative values only)
      Sign will display positive values in green, and the absolutele value of negative values in magenta

    Gamma - changes gamma of display function.  Default value of 1 gives linear mapping of data to pixel value
      pixel values refect value of data^gamma

    Zero Ref - visible for real data, or complex data with "Complex Display" set to R, I, M, or P
               also invisible if "Scalar Display" set to Sign
      This is used for the data-to-pixel value mapping
      --- maps the smallest value to black, and the largest value to white
      0-> maps zero to black, and the largest value to white.  All negative numbers are black
      -0- maps zero to middle gray, with the largest magnitude set to black (if negative) or white (if positive)
      <-0 maps zero to white, and the most negative value to black.  All positive numbers are white

    Fix Range
      If Auto-Range On, the data range for pixel value mapping is rescaled whenever new data appears at input
      If Fixed-Range On, the data range is fixed (and can be changed using Range Min and Range Max)

    Range Min - shows minimum data value used for mapping to pixel values
      This value can be changed if (and only if) "Fix Range" is set to "Fixed-Ranged On"

    Range Max - shows maximum data value used for mapping to pixel values
      This value can be changed if (and only if) "Fix Range" is set to "Fixed-Ranged On"
    """

    def execType(self):
        return gpi.GPI_APPLOOP

    def initUI(self):

        # Widgets
        self.addWidget('ExclusivePushButtons','Complex Display',
                       buttons=['R','I','M','P','C'], val=4)
        self.real_cmaps = ['Gray','IceFire','Fire','Hot','HOT2','BGR']
        self.complex_cmaps = ['HSV','HSL','HUSL','CoolWarm']
        self.addWidget('ExclusivePushButtons','Color Map',
                       buttons=self.real_cmaps, val=0, collapsed=True)
        self.addWidget('SpinBox', 'Edge Pixels', min=0)
        self.addWidget('SpinBox', 'Black Pixels', min=0)
        self.addWidget('DisplayBox', 'Viewport:')
        self.addWidget('Slider', 'Slice', min=1, val=1)
        self.addWidget('ExclusivePushButtons', 'Slice/Tile Dimension', buttons=['0', '1', '2'], val=0)
        self.addWidget('ExclusivePushButtons', 'Extra Dimension', buttons=['Slice', 'Tile', 'RGB(A)'], val=0)
        self.addWidget('SpinBox', '# Columns', val=1)
        self.addWidget('SpinBox', '# Rows', val=1)
        self.addWidget('WindowLevel', 'L W F C:', collapsed=True)
        self.addWidget('ExclusivePushButtons','Scalar Display',
                       buttons=['Pass','Mag','Sign'], val=0)
        self.addWidget('DoubleSpinBox', 'Gamma',min=0.1,max=10,val=1,singlestep=0.05,decimals=3)
        self.addWidget('ExclusivePushButtons','Zero Ref',
                       buttons=['---','0->','-0-','<-0'], val=0)
        self.addWidget('PushButton', 'Fix Range', button_title='Auto-Range On', toggle=True)
        self.addWidget('DoubleSpinBox', 'Range Min')
        self.addWidget('DoubleSpinBox', 'Range Max')

        # IO Ports
        self.addInPort('in', 'NPYarray', drange=(2,3))
        self.addOutPort('out', 'NPYarray')
        self.addOutPort('temp', 'NPYarray')

    def validate(self):

        # Complex or Scalar?
        data = self.getData('in')
        dimfunc = self.getVal('Extra Dimension')

        if data.ndim == 3:
            dimval = self.getVal('Slice/Tile Dimension')
            self.setAttr('Extra Dimension', visible=True)
            if data.shape[-1] not in [3, 4]:
                if dimfunc > 1:
                    dimfunc = 0
                self.setAttr('Extra Dimension', buttons=['Slice', 'Tile'], val=dimfunc)
            else:
                if data.dtype == 'uint8':
                    dimfunc = 2
                self.setAttr('Extra Dimension', buttons=['Slice', 'Tile', 'RGB(A)'], val=dimfunc)

            if dimfunc == 0:
                slval = self.getVal('Slice')
                self.setAttr('Slice/Tile Dimension', visible=True)
                if slval > data.shape[dimval]:
                    slval = data.shape[dimval]
                self.setAttr('Slice', visible=True, min=1, max=data.shape[dimval], val=slval)
                self.setAttr('# Rows', visible=False)
                self.setAttr('# Columns', visible=False)
            elif dimfunc == 1:
                self.setAttr('Slice/Tile Dimension', visible=True)
                ncol = self.getVal('# Columns')
                nrow = self.getVal('# Rows')
                N = data.shape[dimval]

                # set the default to something sane, i.e. square-ish
                if (ncol == 1 and nrow == 1
                    or 'Slice/Tile Dimension' in self.widgetEvents()):
                    ncol = np.round(np.sqrt(N))

                # make sure there are at least enough tiles
                if nrow * ncol < N:
                    nrow = np.ceil(N / ncol)

                # don't add extra blank tiles if they're not needed
                # TODO: same thing, but for columns
                while nrow * ncol - N >= ncol:
                    nrow -= 1

                self.setAttr('# Columns', visible=True, val=ncol)
                self.setAttr('# Rows', visible=True, val=nrow)
                self.setAttr('Slice', visible=False)
            else:
                self.setAttr('Slice/Tile Dimension', visible=False)
                self.setAttr('Slice', visible=False)
                self.setAttr('# Rows', visible=False)
                self.setAttr('# Columns', visible=False)

        else:
            if dimfunc > 1:
                dimfunc = 0
                self.setAttr('Extra Dimension', buttons=['Slice', 'Tile'], val=dimfunc)
            self.setAttr('Extra Dimension', visible=False)
            self.setAttr('Slice/Tile Dimension', visible=False)
            self.setAttr('Slice', visible=False)
            self.setAttr('# Rows', visible=False)
            self.setAttr('# Columns', visible=False)

        self.setAttr('L W F C:',visible=(dimfunc != 2))
        self.setAttr('Gamma',visible=(dimfunc != 2))
        self.setAttr('Fix Range',visible=(dimfunc != 2))

        if dimfunc == 2: # RGBA
          self.setAttr('Complex Display',visible=False)
          self.setAttr('Color Map',visible=False)
          self.setAttr('Scalar Display',visible=False)
          self.setAttr('Edge Pixels',visible=False)
          self.setAttr('Black Pixels',visible=False)
          self.setAttr('Zero Ref',visible=False)
          self.setAttr('Range Min',visible=False)
          self.setAttr('Range Max',visible=False)

        else:

          if np.iscomplexobj(data):
            self.setAttr('Complex Display',visible=True)
            scalarvis = self.getVal('Complex Display') != 4
          else:
            self.setAttr('Complex Display',visible=False)
            scalarvis = True

          if scalarvis:
            self.setAttr('Color Map',buttons=self.real_cmaps,
                         collapsed=self.getAttr('Color Map', 'collapsed'))
          else:
            self.setAttr('Color Map',buttons=self.complex_cmaps,
                         collapsed=self.getAttr('Color Map', 'collapsed'))

          self.setAttr('Scalar Display',visible=scalarvis)
          self.setAttr('Edge Pixels',visible=not scalarvis)
          self.setAttr('Black Pixels',visible=not scalarvis)

          if self.getVal('Scalar Display') == 2:
            self.setAttr('Zero Ref',visible=False)
          else:
            self.setAttr('Zero Ref',visible=scalarvis)

          self.setAttr('Range Min',visible=scalarvis)
          self.setAttr('Range Max',visible=scalarvis)

          zval = self.getVal('Zero Ref')
          if zval == 1:
            self.setAttr('Range Min',val=0)
          elif zval == 3:
            self.setAttr('Range Max',val=0)

          if self.getVal('Fix Range'):
            self.setAttr('Fix Range',button_title="Fixed Range On")
          else:
            self.setAttr('Fix Range',button_title="Auto-Range On")

        return 0

    def compute(self):

        from matplotlib import cm

        # make a copy for changes
        data = self.getData('in').copy()

        # get extra dimension parameters and modify data
        dimfunc = self.getVal('Extra Dimension')
        dimval = self.getVal('Slice/Tile Dimension')
        if data.ndim == 3 and dimfunc < 2:
            if dimfunc == 0: # slice data
                slval = self.getVal('Slice')-1
                if dimval == 0:
                    data = data[slval,...]
                elif dimval == 1:
                    data = data[:,slval,:]
                else:
                    data = data[...,slval]
            else: # tile data
                ncol = self.getVal('# Columns')
                nrow = self.getVal('# Rows')

                # add some blank tiles
                data = np.rollaxis(data, dimval)
                N, xres, yres = data.shape
                N_new = ncol * nrow
                pad_vals = ((0, N_new - N), (0, 0), (0, 0))
                data = np.pad(data, pad_vals, mode='constant')

                # from http://stackoverflow.com/a/13990648/333308
                data = np.reshape(data, (nrow, ncol, xres, yres))
                data = np.swapaxes(data, 1, 2)
                data = np.reshape(data, (nrow*xres, ncol*yres))


        # Read in parameters, make a little floor:ceiling adjustment
        gamma = self.getVal('Gamma')
        lval = self.getAttr('L W F C:', 'val')
        cval = self.getVal('Complex Display')

        if 'Complex Display' in self.widgetEvents():
          if cval == 4:
            self.setAttr('Color Map', buttons=self.complex_cmaps,
                         collapsed=self.getAttr('Color Map', 'collapsed'),
                         val=0)
          # elif self.getAttr('Color Map', 'buttons') != self.real_cmaps:
          # there is no "get_buttons" method, so for now this will reset the
          # colormap whenever "Complex Display" is changed
          # this could/will be added in a future framework update
          else:
            self.setAttr('Color Map', buttons=self.real_cmaps,
                         collapsed=self.getAttr('Color Map', 'collapsed'),
                         val=0)

        cmap = self.getVal('Color Map')
        sval = self.getVal('Scalar Display')
        zval = self.getVal('Zero Ref')
        fval = self.getVal('Fix Range')
        rmin = self.getVal('Range Min')
        rmax = self.getVal('Range Max')

        flor = 0.01*lval['floor']
        ceil = 0.01*lval['ceiling']
        if ceil == flor:
          if ceil == 1.:
            flor = 0.999
          else:
            ceil += 0.001

        # SHOW COMPLEX DATA
        if np.iscomplexobj(data) and cval == 4:
          mag = np.abs(data)
          phase = np.angle(data, deg=True)

          # normalize the mag
          data_min = 0.
          if fval:
            data_max = rmax
          else:
            data_max = mag.max()
            self.setAttr('Range Max',val=data_max)
          data_range = data_max-data_min
          dmask = np.ones(data.shape)
          new_min = data_range*flor + data_min
          new_max = data_range*ceil + data_min
          mag = np.clip(mag, new_min, new_max)

          if new_max > new_min:
            if (gamma == 1): # Put in check for gamma=1, the common use case, just to save time
              mag = (mag - new_min)/(new_max-new_min)
            else:
              mag = pow((mag - new_min)/(new_max-new_min),gamma)
          else:
            mag = np.ones(mag.shape)

          # ADD BORDERS
          edgpix = self.getVal('Edge Pixels')
          blkpix = self.getVal('Black Pixels')
          if (edgpix + blkpix) > 0:
            # new image will be h2 x w2
            # frame defines edge pixels to paint with phase table
            h, w = mag.shape
            h2 = h + 2*(edgpix+blkpix)
            w2 = w + 2*(edgpix+blkpix)
            mag2 = np.zeros((h2,w2))
            phase2 = np.zeros((h2,w2))
            frame = np.zeros((h2,w2)) == 1
            frame[0:edgpix,:] = frame[h2-edgpix:h2,:] = True
            frame[:,0:edgpix] = frame[:,w2-edgpix:w2] = True

            mag2[edgpix+blkpix:edgpix+blkpix+h,edgpix+blkpix:edgpix+blkpix+w] = mag
            mag2[frame] = 1

            phase2[edgpix+blkpix:edgpix+blkpix+h,edgpix+blkpix:edgpix+blkpix+w] = phase
            xloc = np.tile(np.linspace(-1.,1.,w2),(h2,1))
            yloc = np.transpose(np.tile(np.linspace(1.,-1.,h2),(w2,1)))
            phase2[frame] = np.degrees(np.arctan2(yloc[frame],xloc[frame]))

            mag = mag2
            phase = phase2

          # now colorize!
          if cmap == 0: # HSV
            phase_cmap = cm.hsv
          elif cmap == 1: # HSL
            try:
              import seaborn as sns
            except:
              self.log.warn("Seaborn (required for HSL map) not available! Falling back on HSV.")
              phase_cmap = cm.hsv
            else: # from http://stackoverflow.com/a/34557535/333308
              import matplotlib.colors as col
              hlsmap = col.ListedColormap(sns.color_palette("hls", 256))
              phase_cmap = hlsmap
          elif cmap == 2: #HUSL
            try:
              import seaborn as sns
            except:
              self.log.warn("Seaborn (required for HUSL map) not available! Falling back on HSV.")
              phase_cmap = cm.hsv
            else: # from http://stackoverflow.com/a/34557535/333308
              import matplotlib.colors as col
              huslmap = col.ListedColormap(sns.color_palette("husl", 256))
              phase_cmap = huslmap
          elif cmap == 3: # coolwarm
            phase_cmap = cm.coolwarm

          mag_norm = mag
          phase_norm = (phase + 180) / 360
          # phase shift to match old look better
          if cmap != 3:
            phase_norm = (phase_norm - 1/3) % 1
          colorized = 255 * cm.gray(mag_norm) * phase_cmap(phase_norm)
          red = colorized[...,0]
          green = colorized[...,1]
          blue = colorized[...,2]
          alpha = colorized[...,3]

        # DISPLAY SCALAR DATA
        elif dimfunc != 2:

          if np.iscomplexobj(data):
            if cval == 0: # Real
              data = np.real(data)
            elif cval == 1: # Imag
              data = np.imag(data)
            elif cval == 2: # Mag
              data = np.abs(data)
            elif cval == 3: # Phase
              data = np.angle(data, deg=True)

          if sval == 1: # Mag
            data = np.abs(data)
          elif sval == 2: # Sign
            sign = np.sign(data)
            data = np.abs(data)

          # normalize the data
          if fval:
            data_min = rmin
            data_max = rmax
          else:
            data_min = data.min()
            data_max = data.max()

          if sval != 2:
            if zval == 1:
              data_min = 0.
            elif zval == 2:
              data_max = max(abs(data_min),abs(data_max))
              data_min = -data_max
            elif zval == 3:
              data_max = 0.
            data_range = data_max-data_min
            self.setAttr('Range Min',val=data_min)
            self.setAttr('Range Max',val=data_max)
          else:
            data_min = 0.
            data_max = max(abs(data_min),abs(data_max))
            data_range = data_max
            self.setAttr('Range Min',val=-data_range)
            self.setAttr('Range Max',val=data_range)

          dmask = np.ones(data.shape)
          new_min = data_range*flor + data_min
          new_max = data_range*ceil + data_min
          data = np.minimum(np.maximum(data,new_min*dmask),new_max*dmask)

          if new_max > new_min:
            if (gamma == 1): # Put in check for gamma=1, the common use case, just to save time
              data = 255.*(data - new_min)/(new_max-new_min)
            else:
              data = 255.*pow((data - new_min)/(new_max-new_min),gamma)
          else:
            data = 255.*np.ones(data.shape)

          if sval != 2: #Not Signed Data (Pass or Mag)
            # Show based on a color map
            if cmap == 0: # Grayscale
              red = green = blue = np.uint8(data)
              alpha = 255. * np.ones(blue.shape)
            else:
              rd = np.zeros(data.shape)
              gn = np.zeros(data.shape)
              be = np.zeros(data.shape)
              zmask = np.zeros(data.shape)
              fmask = np.ones(data.shape)

              if cmap == 1: # IceFire
                hue = 4.*(data/256.)
                hindex0 =                          hue < 1.
                hindex1 = np.logical_and(hue >= 1.,hue < 2.)
                hindex2 = np.logical_and(hue >= 2.,hue < 3.)
                hindex3 = np.logical_and(hue >= 3.,hue < 4.)

                be[hindex0] = hue[hindex0]
                gn[hindex0] = zmask[hindex0]
                rd[hindex0] = zmask[hindex0]

                gn[hindex1] = (hue-1.)[hindex1]
                rd[hindex1] = (hue-1.)[hindex1]
                be[hindex1] = fmask[hindex1]

                gn[hindex2] = fmask[hindex2]
                rd[hindex2] = fmask[hindex2]
                be[hindex2] = (3.-hue)[hindex2]

                rd[hindex3] = fmask[hindex3]
                gn[hindex3] = (4.-hue)[hindex3]
                be[hindex3] = zmask[hindex3]

              elif cmap == 2: # Fire
                hue = 4.*(data/256.)
                hindex0 =                          hue < 1.
                hindex1 = np.logical_and(hue >= 1.,hue < 2.)
                hindex2 = np.logical_and(hue >= 2.,hue < 3.)
                hindex3 = np.logical_and(hue >= 3.,hue < 4.)

                be[hindex0] = hue[hindex0]
                rd[hindex0] = zmask[hindex0]
                gn[hindex0] = zmask[hindex0]

                be[hindex1] = (2.-hue)[hindex1]
                rd[hindex1] = (hue-1.)[hindex1]
                gn[hindex1] = zmask[hindex1]

                rd[hindex2] = fmask[hindex2]
                gn[hindex2] = (hue-2.)[hindex2]
                be[hindex2] = zmask[hindex2]

                rd[hindex3] = fmask[hindex3]
                gn[hindex3] = fmask[hindex3]
                be[hindex3] = (hue-3.)[hindex3]

              elif cmap == 3: # Hot
                hue = 3.*(data/256.)
                hindex0 =                          hue < 1.
                hindex1 = np.logical_and(hue >= 1.,hue < 2.)
                hindex2 = np.logical_and(hue >= 2.,hue < 3.)

                rd[hindex0] = hue[hindex0]
                be[hindex0] = zmask[hindex0]
                gn[hindex0] = zmask[hindex0]

                gn[hindex1] = (hue-1.)[hindex1]
                rd[hindex1] = fmask[hindex1]
                be[hindex1] = zmask[hindex1]

                rd[hindex2] = fmask[hindex2]
                gn[hindex2] = fmask[hindex2]
                be[hindex2] = (hue-2.)[hindex2]

              if cmap == 4: # Hot2, from ASIST (http://asist.umin.jp/index-e.htm)
                rindex0 = data < 20.0
                rindex1 = np.logical_and(data >=  20.0, data <= 100.0)
                rindex2 = np.logical_and(data > 100.0, data < 128.0)
                rindex3 = np.logical_and(data >= 128.0, data <= 191.0)
                rindex4 = data > 191.0
                rd[rindex0] = data[rindex0] * 4.0
                rd[rindex1] = 80.0 - (data[rindex1] - 20.0)
                rd[rindex3] = (data[rindex3] - 128.0) * 4.0
                rd[rindex4] = data[rindex4] * 0.0 + 255.0
                rd = rd/255.0;

                gindex0 = data < 45.0
                gindex1 = np.logical_and(data >= 45.0, data <= 130.0)
                gindex2 = np.logical_and(data > 130.0, data < 192.0)
                gindex3 = data >= 192.0
                gn[gindex1] = (data[gindex1] - 45.0)*3.0
                gn[gindex2] = data[gindex2] * 0.0 + 255.0
                gn[gindex3] = 252.0 - (data[gindex3] - 192.0)*4.0
                gn = gn/255.0

                bindex0 = (data < 1.0)
                bindex1 = np.logical_and(data >= 1.0, data < 86.0)
                bindex2 = np.logical_and(data >= 86.0, data <= 137.0)
                bindex3 = data > 137.0
                be[bindex1] = (data[bindex1] - 1.0)*3.0
                be[bindex2] = 255.0 - (data[bindex2] - 86.0)*5.0
                be = be/255.0

              elif cmap == 5: # BGR
                hue = 4.*(data/256.)
                hindex0 =                          hue < 1.
                hindex1 = np.logical_and(hue >= 1.,hue < 2.)
                hindex2 = np.logical_and(hue >= 2.,hue < 3.)
                hindex3 = np.logical_and(hue >= 3.,hue < 4.)

                be[hindex0] = hue[hindex0]
                gn[hindex0] = zmask[hindex0]
                rd[hindex0] = zmask[hindex0]

                gn[hindex1] = (hue-1.)[hindex1]
                rd[hindex1] = zmask[hindex1]
                be[hindex1] = fmask[hindex1]

                gn[hindex2] = fmask[hindex2]
                rd[hindex2] = (hue-2.)[hindex2]
                be[hindex2] = (3.-hue)[hindex2]

                rd[hindex3] = fmask[hindex3]
                gn[hindex3] = (4.-hue)[hindex3]
                be[hindex3] = zmask[hindex3]

              blue = np.uint8(255.*rd)
              red = np.uint8(255.*be)
              green = np.uint8(255.*gn)
              alpha = np.uint8(255.*np.ones(blue.shape))

          else: #Signed data, positive numbers green, negative numbers magenta
            red = np.zeros(data.shape)
            green = np.zeros(data.shape)
            blue = np.zeros(data.shape)
            red[sign<=0] = data[sign<=0]
            blue[sign<=0] = data[sign<=0]
            green[sign>=0] = data[sign>=0]

            red = red.astype(np.uint8)
            green = green.astype(np.uint8)
            blue = blue.astype(np.uint8)
            alpha = np.uint8(data)

        # DISPLAY RGB image
        else:

          if data.shape[-1] > 3:
            red   = data[:,:,0].astype(np.uint8)
            green = data[:,:,1].astype(np.uint8)
            blue  = data[:,:,2].astype(np.uint8)
            if(data.ndim == 3 and data.shape[-1] == 4) :
                alpha = data[:,:,3].astype(np.uint8)
            else:
                alpha = 255.*np.ones(blue.shape)
          else:
              self.log.warn("input veclen of "+str(data.shape[-1])+" is incompatible")
              return 1

        h, w = red.shape[:2]
        image1 = np.zeros((h, w, 4), dtype=np.uint8)
        image1[:, :, 0] = red
        image1[:, :, 1] = green
        image1[:, :, 2] = blue
        image1[:, :, 3] = alpha
        format_ = QtGui.QImage.Format_RGB32

        image = QtGui.QImage(image1.data, w, h, format_)

        #send the RGB values to the output port
        imageTru = np.zeros((h, w, 4), dtype=np.uint8)
        imageTru[:, :, 0] = red
        imageTru[:, :, 1] = green
        imageTru[:, :, 2] = blue
        imageTru[:, :, 3] = alpha
        image.ndarray = imageTru
        if image.isNull():
            self.log.warn("Image Viewer: cannot load image")

        self.setAttr('Viewport:', val=image)
        self.setData('out',imageTru)

        return 0



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: David Smith
# Date: 2013Nov22

import gpi
import numpy as np


class ExternalNode(gpi.NodeAPI):
    """
        Generates a Shepp-Logan phantom of the designated size.
    """
    def execType(self):
        return gpi.GPI_PROCESS

    def initUI(self):

        # Widgets
        self.addWidget('TextBox', 'Info')
        self.addWidget('SpinBox', 'Size', val=128, min=1, max=8192)
        self.addWidget('PushButton','Flip', toggle=True, val=True)
        self.addWidget('PushButton','Compute', toggle=True, val=True)
        self.addWidget('SpinBox', 'Bandlimit iterations', val=0, min=0)

        # IO Ports
        self.addOutPort('out', 'NPYarray')

    def validate(self):

        return 0


    def compute(self):

        #from phantom import phantom
        itr = self.getVal('Bandlimit iterations')

        # visibility
        if self.getVal('Compute'):

            n = self.getVal('Size')
            out = phantom(n)

            out = self.condition(out, itr)
            if self.getVal('Flip'):
                out = out[::-1]

            self.setData('out', out)

        return(0)

    def condition(self, data, iter=0):
        import numpy as np

        # don't condition
        if iter == 0:
            return data

        # recast to fftw reqs.
        orig_type = data.dtype
        data = data.astype(np.complex64, copy=False)

        band = self.window2(data.shape, windowpct=10, widthpct=100)

        # band limit in both domains
        for i in range(iter):
            data *= band # image space
            data = self.fft2(data, dir=0)
            data *= band # k-space
            data = self.fft2(data, dir=1)

        # keep original type
        data = np.abs(data).astype(orig_type, copy=False)

        return data

    def fft2(self, data, dir=0, zp=1, out_shape=[], tx_ON=True):
        # data: np.complex64
        # dir: int (0 or 1)
        # zp: float (>1)

        # simplify the fftw wrapper
        import numpy as np
        import gpi_core.math.fft as corefft

        # generate output dim size array
        # fortran dimension ordering
        outdims = list(data.shape)
        if len(out_shape):
            outdims = out_shape
        else:
            for i in range(len(outdims)):
                outdims[i] = int(outdims[i]*zp)
        outdims.reverse()
        outdims = np.array(outdims, dtype=np.int64)

        # load fft arguments
        kwargs = {}
        kwargs['dir'] = dir

        # transform or just zeropad
        if tx_ON:
            kwargs['dim1'] = 1
            kwargs['dim2'] = 1
        else:
            kwargs['dim1'] = 0
            kwargs['dim2'] = 0

        return corefft.fftw(data, outdims, **kwargs)

    def window2(self, shape, windowpct=100.0, widthpct=100.0, stopVal=0, passVal=1):
        # 2D hanning window just like shapes
        #   OUTPUT: 2D float32 circularly symmetric hanning

        import numpy as np

        # window function width
        bnd = 100.0/widthpct

        # generate coords for each dimension
        x = np.linspace(-bnd, bnd, shape[-1], endpoint=(shape[-1] % 2 != 0))
        y = np.linspace(-bnd, bnd, shape[-2], endpoint=(shape[-2] % 2 != 0))

        # create a 2D grid with coordinates then get radial coords
        xx, yy = np.meshgrid(x,y)
        radius = np.sqrt(xx*xx + yy*yy)

        # calculate hanning
        windIdx = radius <= 1.0
        passIdx = radius <= (1.0 - (windowpct/100.0))
        func = 0.5 * (1.0 - np.cos(np.pi * (1.0 - radius[windIdx]) / (windowpct/100.0)))

        # populate output array
        out = np.zeros(shape, dtype=np.float32)
        out[windIdx] = stopVal + func * (passVal - stopVal)
        out[passIdx] = passVal

        return out

## Copyright (C) 2010  Alex Opie  <lx_op@orcon.net.nz>
##
## This program is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or (at
## your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; see the file COPYING.  If not, see
## <http://www.gnu.org/licenses/>.

import numpy as np

def phantom (n = 256, p_type = 'Modified Shepp-Logan', ellipses = None):
	"""
	 phantom (n = 256, p_type = 'Modified Shepp-Logan', ellipses = None)

	Create a Shepp-Logan or modified Shepp-Logan phantom.

	A phantom is a known object (either real or purely mathematical)
	that is used for testing image reconstruction algorithms.  The
	Shepp-Logan phantom is a popular mathematical model of a cranial
	slice, made up of a set of ellipses.  This allows rigorous
	testing of computed tomography (CT) algorithms as it can be
	analytically transformed with the radon transform (see the
	function `radon').

	Inputs
	------
	n : The edge length of the square image to be produced.

	p_type : The type of phantom to produce. Either
	  "Modified Shepp-Logan" or "Shepp-Logan".  This is overridden
	  if `ellipses' is also specified.

	ellipses : Custom set of ellipses to use.  These should be in
	  the form
	  	[[I, a, b, x0, y0, phi],
	  	 [I, a, b, x0, y0, phi],
	  	 ...]
	  where each row defines an ellipse.
	  I : Additive intensity of the ellipse.
	  a : Length of the major axis.
	  b : Length of the minor axis.
	  x0 : Horizontal offset of the centre of the ellipse.
	  y0 : Vertical offset of the centre of the ellipse.
	  phi : Counterclockwise rotation of the ellipse in degrees,
	        measured as the angle between the horizontal axis and
	        the ellipse major axis.
	  The image bounding box in the algorithm is [-1, -1], [1, 1],
	  so the values of a, b, x0, y0 should all be specified with
	  respect to this box.

	Output
	------
	P : A phantom image.

	Usage example
	-------------
	  import matplotlib.pyplot as pl
	  P = phantom ()
	  pl.imshow (P)

	References
	----------
	Shepp, L. A.; Logan, B. F.; Reconstructing Interior Head Tissue
	from X-Ray Transmissions, IEEE Transactions on Nuclear Science,
	Feb. 1974, p. 232.

	Toft, P.; "The Radon Transform - Theory and Implementation",
	Ph.D. thesis, Department of Mathematical Modelling, Technical
	University of Denmark, June 1996.

	"""

	if (ellipses is None):
		ellipses = _select_phantom (p_type)
	elif (np.size (ellipses, 1) != 6):
		raise AssertionError ("Wrong number of columns in user phantom")

	# Blank image
	p = np.zeros ((n, n))

	# Create the pixel grid
	ygrid, xgrid = np.mgrid[-1:1:(1j*n), -1:1:(1j*n)]

	for ellip in ellipses:
		I   = ellip [0]
		a2  = ellip [1]**2
		b2  = ellip [2]**2
		x0  = ellip [3]
		y0  = ellip [4]
		phi = ellip [5] * np.pi / 180  # Rotation angle in radians

		# Create the offset x and y values for the grid
		x = xgrid - x0
		y = ygrid - y0

		cos_p = np.cos (phi)
		sin_p = np.sin (phi)

		# Find the pixels within the ellipse
		locs = (((x * cos_p + y * sin_p)**2) / a2
              + ((y * cos_p - x * sin_p)**2) / b2) <= 1

		# Add the ellipse intensity to those pixels
		p [locs] += I

	return p


def _select_phantom (name):
	if (name.lower () == 'shepp-logan'):
		e = _shepp_logan ()
	elif (name.lower () == 'modified shepp-logan'):
		e = _mod_shepp_logan ()
	else:
		raise ValueError ("Unknown phantom type: %s" % name)

	return e


def _shepp_logan ():
	#  Standard head phantom, taken from Shepp & Logan
	return [[   2,   .69,   .92,    0,      0,   0],
	        [-.98, .6624, .8740,    0, -.0184,   0],
	        [-.02, .1100, .3100,  .22,      0, -18],
	        [-.02, .1600, .4100, -.22,      0,  18],
	        [ .01, .2100, .2500,    0,    .35,   0],
	        [ .01, .0460, .0460,    0,     .1,   0],
	        [ .02, .0460, .0460,    0,    -.1,   0],
	        [ .01, .0460, .0230, -.08,  -.605,   0],
	        [ .01, .0230, .0230,    0,  -.606,   0],
	        [ .01, .0230, .0460,  .06,  -.605,   0]]

def _mod_shepp_logan ():
	#  Modified version of Shepp & Logan's head phantom,
	#  adjusted to improve contrast.  Taken from Toft.
	return [[   1,   .69,   .92,    0,      0,   0],
	        [-.80, .6624, .8740,    0, -.0184,   0],
	        [-.20, .1100, .3100,  .22,      0, -18],
	        [-.20, .1600, .4100, -.22,      0,  18],
	        [ .10, .2100, .2500,    0,    .35,   0],
	        [ .10, .0460, .0460,    0,     .1,   0],
	        [ .10, .0460, .0460,    0,    -.1,   0],
	        [ .10, .0460, .0230, -.08,  -.605,   0],
	        [ .10, .0230, .0230,    0,  -.606,   0],
	        [ .10, .0230, .0460,  .06,  -.605,   0]]

#def ?? ():
#	# Add any further phantoms of interest here
#	return np.array (
#	 [[ 0, 0, 0, 0, 0, 0],
#	  [ 0, 0, 0, 0, 0, 0]])




# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Sudarshan Ragunathan
# Date: 2014may28

import numpy as np
import gpi


class ExternalNode(gpi.NodeAPI):
    """Performs a 1D-histogram of the non-zero values in a Numpy Array.

    INPUT - numpy array (data type : float or complex)
    
    OUTPUT:
    HistVal - Values in each histogram bin (1D numpy array)
    HistBin - Vector of bins (1D numpy array)

    WIDGETS:
    R I M - Choose from Real, Imaginary or Magnitude data (Visible for complex input)
    Bins - Number of bins
    Range - Lower and Upper range of the bins. The user has the option of specifying
            the range or using the max and min of the input
    """

    def initUI(self):

        # Widgets
        self.addWidget('ExclusivePushButtons', 'R I M', buttons=['R', 'I', 'M'], val=2)
        self.addWidget('DoubleSpinBox', '# Bins', val=500)
        self.addWidget('DoubleSpinBox', 'Lower Bound', val=0.)
        self.addWidget('DoubleSpinBox', 'Upper Bound', val=255.)
        self.bin_op = ['Set Range','Use Min/Max']
        self.addWidget('ExclusiveRadioButtons', 'Bin Range', buttons=self.bin_op, val=0)

        # IO Ports
        self.addInPort('Data In', 'NPYarray', obligation=gpi.REQUIRED)
        self.addOutPort('HistVal', 'NPYarray')
        self.addOutPort('HistBin', 'NPYarray')

    def validate(self):
        im_input = self.getData('Data In')
        
        # Check for dtype of Input (Visibility of R I M op)
        im_type = str(im_input.dtype)
        if 'float' in im_type:
            self.setAttr('R I M', visible=False)
            data_in = im_input
        elif 'complex' in im_type:
            self.setAttr('R I M', visible=True)
            op = self.getVal('R I M')
            if op == 0:
                data_in = np.real(im_input)
            elif op == 1:
                data_in = np.imag(im_input)
            else:
                data_in = np.abs(im_input)
        data = data_in[np.nonzero(data_in)] 
        data_min = np.amin(data)
        data_max = np.amax(data)

        # Reset the min and max of the lower and upper bounds of bins
        #self.setAttr('Lower Bound', val=data_min)
        #self.setAttr('Upper Bound', val=data_max)
        bin_range = self.getVal('Bin Range')
        if bin_range == 1:
            self.setAttr('Lower Bound', visible=False)
            self.setAttr('Upper Bound', visible=False)
            self.setAttr('Lower Bound', val=data_min)
            self.setAttr('Upper Bound', val=data_max)
        else:
            self.setAttr('Lower Bound', visible=True)
            self.setAttr('Upper Bound', visible=True)

        # Check for Lower > Upper Bound
        lbound = self.getVal('Lower Bound')
        ubound = self.getVal('Upper Bound')
        if lbound > ubound:
            self.setAttr('Lower Bound', val=np.maximum(lbound,ubound))
            self.setAttr('Upper Bound', val=np.maximum(lbound,ubound))

        return(0)

    def compute(self):

        import numpy as np
        
        im_input = self.getData('Data In')
        op = self.getVal('R I M')
        if op == 0:
            data_in = np.real(im_input)
        elif op == 1:
            data_in = np.imag(im_input)
        else:
            data_in = np.abs(im_input)
        data = data_in[np.nonzero(data_in)]
        lbound = self.getVal('Lower Bound')
        ubound = self.getVal('Upper Bound')
        bins = self.getVal('# Bins')

        hist_out = np.histogram(data,bins,range=(lbound,ubound))
        self.setData('HistVal', hist_out[0])
        self.setData('HistBin', hist_out[1])
        return(0)







# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.

# Author: Nick Zwart
# Date: 2013 sep 24

import math
import gpi
from gpi.numpyqt import numpy2qimage
from gpi import QtCore, QtWidgets

import numpy as np
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import (
    FigureCanvas, NavigationToolbar2QT as NavigationToolbar)

from scipy import ndimage


class MatplotDisplay2(gpi.GenericWidgetGroup):
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        super().__init__(title, parent)

        self._data = None
        self.create_main_frame()
        self.on_draw()

    # setters
    def set_val(self, data):
        '''Takes a list of npy arrays.
        '''
        if isinstance(data, list):
            self._data = data
            self.on_draw()
        else:
            return

    # getters
    def get_val(self):
        return self._data

    # support
    def create_main_frame(self):

        self.fig = Figure((5.0, 4.0), dpi=100)
        self.canvas = FigureCanvas(self.fig)
        self.canvas.setParent(self)
        self.canvas.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.canvas.setFocus()

        self.mpl_toolbar = NavigationToolbar(self.canvas, self)

        self.canvas.mpl_connect('key_press_event', self.on_key_press)

        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(self.canvas)  # the matplotlib canvas
        vbox.addWidget(self.mpl_toolbar)
        self.setLayout(vbox)

    def on_draw(self):
        self.fig.clear()
        self.axes = self.fig.add_subplot(111)
        # self.axes.plot(self.x, self.y, 'ro')
        # self.axes.imshow(self.data, interpolation='nearest')
        # self.axes.plot([1,2,3])

        if self._data is None:
            return

        # plot each set
        # print "--------------------plot the data"
        for data in self._data:

            # check for x, y data
            if data.shape[-1] == 2:
                self.axes.plot(data[..., 0], data[..., 1], alpha=0.8, lw=2.0)
            else:
                self.axes.plot(data, alpha=0.8, lw=2.0)

        self.canvas.draw()

    def on_key_press(self, event):
        # print 'Matplotlib-> you pressed:' + str(event.key)
        # implement the default mpl key press events described at
        # http://matplotlib.org/users/navigation_toolbar.html#navigation-
        # keyboard-shortcuts
        try:
            from matplotlib.backend_bases import key_press_handler
            key_press_handler(event, self.canvas, self.mpl_toolbar)
        except:
            print("key_press_handler import failed. -old matplotlib version.")


# WIDGET
class WindowLevel(gpi.GenericWidgetGroup):

    """Provides an interface to the BasicCWFCSliders."""
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        super().__init__(title, parent)
        self.sl = gpi.BasicCWFCSliders()
        self.sl.valueChanged.connect(self.valueChanged)
        self.pb = gpi.BasicPushButton()
        self.pb.set_button_title('reset')
        # layout
        wdgLayout = QtWidgets.QVBoxLayout()
        wdgLayout.addWidget(self.sl)
        wdgLayout.addWidget(self.pb)
        self.setLayout(wdgLayout)
        # default
        self.set_min(0)
        self.set_max(100)
        self.sl.set_allvisible(True)
        self.reset_sliders()
        self.pb.valueChanged.connect(self.reset_sliders)

    # setters
    def set_val(self, val):
        """Set multiple values with a python-dict with keys:
        level, window, floor and ceiling. -Requires integer
        values for each key.
        """
        self.sl.set_center(val['level'])
        self.sl.set_width(val['window'])
        self.sl.set_floor(val['floor'])
        self.sl.set_ceiling(val['ceiling'])

    def set_min(self, val):
        """Set min for level, window, floor and ceiling (int)."""
        self.sl.set_min(val)

    def set_max(self, val):
        """Set max for level, window, floor and ceiling (int)."""
        self.sl.set_max(val)

    # getters
    def get_val(self):
        val = {}
        val['level'] = self.sl.get_center()
        val['window'] = self.sl.get_width()
        val['floor'] = self.sl.get_floor()
        val['ceiling'] = self.sl.get_ceiling()
        return val

    def get_min(self):
        return self.sl.get_min()

    def get_max(self):
        return self.sl.get_max()

    def reset_sliders(self):
        val = {}
        val['window'] = 100
        val['level'] = 50
        val['floor'] = 0
        val['ceiling'] = 100
        self.set_val(val)


class ExternalNode(gpi.NodeAPI):

    """Display image of 2D array.  Dragging across image with left mouse button produces a graph of signal value along that line.

    INPUT:
    2D array

    WIDGETS:
    Viewport - displays image
      Double clicking on Viewport area brings up a scaling widget to change the image size, and change graphic overlay

    L W F C - (hidden by default - double click on widget area to show sliders)
              Adjust value-to-pixel brightness mapping using Level/Window or Floor/Ceiling
    Cross Section - hidden until line is drawn on image, then a graph of data values along line
    """

    def execType(self):
        return gpi.GPI_APPLOOP

    def initUI(self):

        # Widgets
        self.addWidget('DisplayBox', 'Viewport:')
        self.addWidget('WindowLevel', 'L W F C:', collapsed=True)
        self.addWidget('MatplotDisplay2', 'Cross Section', visible=False)

        # IO Ports
        self.addInPort('in', 'NPYarray', ndim=2, obligation=gpi.REQUIRED)

    def compute(self):

        # make a copy for changes
        data = self.getData('in').copy()

        # convert complex to mag
        if np.iscomplexobj(data):
            data = np.abs(data)

        # get rid of negative numbers
        if data.min() < 0.:
            data -= data.min()

        # normalize the data
        data_min = data.min()
        data_max = data.max()
        data_range = data_max - data_min

        val = self.getAttr('L W F C:', 'val')

        new_min = data_range * val['floor'] / 100.0 + data_min
        new_max = data_range * val['ceiling'] / 100.0 + data_min

        data[data < new_min] = new_min
        data[data > new_max] = new_max

        data = data - math.fabs(new_min)
        data = data / (new_max - math.fabs(new_min)) * 255

        image = numpy2qimage(data.astype(np.uint8))
        if image.isNull():
            self.log.warn("Image Viewer: cannot load image")
        else:
            self.setAttr('Viewport:', val=image)

        line = self.getAttr('Viewport:', 'line')
        if line:
            a = self.getData('in')
            if np.iscomplexobj(a):
                a = np.abs(a)

            # Make a line with "l" points
            x0, y0 = line[0]
            x1, y1 = line[1]
            l = int(np.hypot(x1 - x0, y1 - y0))
            x, y = np.linspace(x0, x1, l), np.linspace(y0, y1, l)

            # Extract the values along the line, using cubic interpolation
            arr = ndimage.map_coordinates(a, np.vstack((x, y)), order=3)

            self.setAttr('Cross Section', val=[arr])
            self.setAttr('Cross Section', visible=True)
        else:
            self.setAttr('Cross Section', visible=False)

        return(0)



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Payal Bhavsar
# Date: 2013Nov18

import gpi

class ExternalNode(gpi.NodeAPI):
    """A module for writing NPY arrays to a comma separated value list in ASCII text.

    INPUT: Numpy array to write to file

    WIDGETS:
    File Browser - button to launch file browser, and typein widget if the pathway is known.
    Write Mode - write at any event, or write only with new filename
    Write Now - write right now
    """

    def initUI(self):
       # Widgets
        self.addWidget(
            'SaveFileBrowser', 'File Browser', button_title='Browse',
            caption='Save File (*.csv)', filter='Comma Seperated Values (*.csv)')
        self.addWidget('PushButton', 'Write Mode', button_title='Write on New Filename', toggle=True)
        self.addWidget('PushButton', 'Write Now', button_title='Write Right Now', toggle=False)

        # IO Ports
        self.addInPort('in', 'NPYarray')

        self.URI = gpi.TranslateFileURI

    def validate(self):
        if self.getVal('Write Mode'):
            self.setAttr('Write Mode', button_title="Write on Every Event")
        else:
            self.setAttr('Write Mode', button_title="Write on New Filename")

        fname = self.URI(self.getVal('File Browser'))
        self.setDetailLabel(fname)

        return 0

    def compute(self):
        import numpy as np
        import csv

        if self.getVal('Write Mode') or self.getVal('Write Now') or ('File Browser'  in self.widgetEvents()):
            fname = self.URI(self.getVal('File Browser'))
            if not fname.endswith('.csv'):
                fname += '.csv'
            if fname == '.csv':
                return 0

            # get data and check it
            data = self.getData('in')
            if data.dtype in [np.complex, np.complex64, np.complex128, np.complex256]:
                self.log.warn('Complex data is not readable by most CSV readers including the ReadCSV node.\n\tYou can split your real and imag sets into another dimension if needed.')
            if len(data.shape) > 2:
                self.log.warn('ndim > 2 not supported')
                return 1

            np.savetxt(fname,data,delimiter=",")

        return(0)



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pipe
# Date: 2013Jun18

import gpi
import numpy as np

class ExternalNode(gpi.NodeAPI):
    """Reformats pulse sequence information from  Bloch- (potentially cascaded
    Bloch-) node(s) for plotting in 1D viewer/plotter nodes.
    """

    def initUI(self):
        # Widgets

        # IO Ports
        self.addInPort('M_in', 'NPYarray',obligation=gpi.REQUIRED)
        self.addOutPort('B_out', 'NPYarray')

        return 0

    def validate(self):
        return 0

    def compute(self):

        # New data
        m_in = self.getData('M_in')
        mdim = list(m_in.shape)

# X Axis
        xval = m_in[6,slice(1,mdim[1]),0,0,0,0,0,0,0,0,0]
# Y Axes
        outvals = m_in[:,slice(1,mdim[1]),0,0,0,0,0,0,0,0,0]
        zr = np.zeros(mdim[1]-1)
        gx = outvals[7,:]
        gy = outvals[8,:]
        gz = outvals[9,:]
        rfx = 1000.*outvals[10,:] # in units of uT
        rfy = 1000.*outvals[11,:] # in units of uT
        y0 = np.column_stack((zr,gx,gy,gz,rfx,rfy))
        x0 = np.column_stack((xval,xval,xval,xval,xval,xval))

        b_out = np.append(x0[:,:,np.newaxis],y0[:,:,np.newaxis],axis=2)

        self.setData('B_out', b_out)

        return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pipe
# Date: 2013Jul

import numpy as np
import scipy
from scipy import integrate as int
import gpi

class ExternalNode(gpi.NodeAPI):
    """Integrate or Differentiate along a given axis

    INPUT - numpy array to operate on

    OUTPUT - numpy array after operation

    WIDGETS:
    Operation - select Integerate or Differentiate operation
    Dimension - select axis along with to integerate or differentiate
    """

    def initUI(self):

        # Widgets
        # Edge clause default setting
        self.edgeMode = 0
        self.func = 0
        self.op_labels = ['Integrate', 'Derivative','Discrete Diff.']
        self.addWidget('ExclusivePushButtons', 'Operation', buttons=self.op_labels
                        , val = 0)
        self.addWidget('Slider', 'Dimension', val = 0, min=0)
        self.addWidget('ExclusivePushButtons', 'Endpoint Treatment', buttons = [
                       '1st order','2nd order'], val = self.edgeMode)
        self.addWidget('ExclusivePushButtons', 'Method', buttons = [
                       'Rect.','Trap.'], val = self.edgeMode)

        # IO Ports
        self.addInPort('in', 'NPYarray', obligation=gpi.OPTIONAL)
        self.addOutPort('out', 'NPYarray')

    def validate(self):
        data = self.getData('in')

        if self.getVal('Operation') == 0:
            self.setAttr('Endpoint Treatment', visible = False)
            self.setAttr('Method', visible = True)
        elif self.getVal('Operation') == 1:
            self.setAttr('Endpoint Treatment', visible = True)
            self.setAttr('Method', visible = False)
        else:
            self.setAttr('Endpoint Treatment', visible = False)
            self.setAttr('Method', visible = False)

        if data is not None:
            self.setAttr('Dimension', max=(len(data.shape)-1), min=(-len(data.shape)))

        self.setDetailLabel('{}'.format(self.op_labels[self.getVal('Operation')]))
        return(0)

    def compute(self):
        import numpy as np

        data = self.getData('in')
        if data is not None:
            if self.getVal('Operation') == 0:
                if self.getVal('Method') == 0:
                    out = np.cumsum(data, axis=self.getVal('Dimension'))
                else:
                    out = int.cumtrapz(data, axis=self.getVal('Dimension'),
                                        initial = 0)
            elif self.getVal('Operation') == 1:
                order = self.getVal('Endpoint Treatment') + 1;
                out = np.gradient(data, axis=self.getVal('Dimension'),
                                edge_order=order)
            else:
                out = np.diff(data,axis=self.getVal('Dimension'))
            self.setData('out', out)

        return(0)

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart
# Date: 2012apr03

import os
import time
import gpi
import pickle as pic

class ExternalNode(gpi.NodeAPI):
    """Provides an interface to the python pickle (cPickle) module for
    de-serializing py-objects from a file.

    OUTPUT: Numpy array read from file

    WIDGETS:
    I/O Info - Gives info on data file and data type
    File Browser - button to launch file browser, and typein widget if the pathway is known.
    """
    def execType(self):
        # default executable type
        #return gpi.GPI_THREAD
        return gpi.GPI_PROCESS # this is the safest

    def initUI(self):

       # Widgets
        self.addWidget('TextBox', 'I/O Info:')
        self.addWidget('OpenFileBrowser', 'File Browser',
                button_title='Browse', caption='Open File',
                    filter='pickled (*.pickle);;all (*)')

        # IO Ports
        self.addOutPort('out','PASS')

        # store for later use
        self.URI = gpi.TranslateFileURI

    def validate(self):
        fname = self.URI(self.getVal('File Browser'))
        self.setDetailLabel(fname)

    def compute(self):

        import os
        import time
        import pickle as pic

        # start file browser
        fname = self.URI(self.getVal('File Browser'))

        # check that the path actually exists
        if not os.path.exists(fname):
            self.log.node("Path does not exist: "+str(fname))
            return 0

        # show some file stats
        fstats = os.stat(fname)
        # creation
        ctime = time.strftime('%d/%m/20%y', time.localtime(fstats.st_ctime))
        # mod time
        mtime = time.strftime('%d/%m/20%y', time.localtime(fstats.st_mtime))
        # access time
        atime = time.strftime('%d/%m/20%y', time.localtime(fstats.st_atime))
        # filesize
        fsize = fstats.st_size
        # user id
        uid = fstats.st_uid
        # group id
        gid = fstats.st_gid

        # read the data
        fptr = open(fname, "rb")
        out = pic.load(fptr)
        fptr.close()

        info = "created: "+str(ctime)+"\n" \
               "accessed: "+str(atime)+"\n" \
               "modified: "+str(mtime)+"\n" \
               "UID: "+str(uid)+"\n" \
               "GID: "+str(gid)+"\n" \
               "file size (bytes): "+str(fsize)+"\n" \
               "type: "+str(type(out))+"\n"
        self.setAttr('I/O Info:', val=info)

        self.setData('out', out)

        return(0)



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pipe
# Date: 2013Jun18

import gpi
import numpy as np

class ExternalNode(gpi.NodeAPI):
    """Reformats magnetization profiles defined in Spyn and Bloch nodes for
    plotting in 1D viewer/plotter nodes.

    INPUT: data from Bloch or Spyn node

    OUTPUT: data to plot (typically via Matplotlib)

    WIDGETS:
    M display - select which values to display
                (Mx, My, Mz, Mt=sqrt(mx^2+my^2), |M| = sqrt(mx^2+my^2+mz^2)
                selecting multiple buttons will produce multiple graphs
    Average Across: for each dimension (e.g. T1, T2, etc.), the magnetization can be averaged across that dimension
    X Axis - the independent variable by which to plot on the abscissa
             (cannot be a dimension that has been averaged across)
    (T,T1,T2,FQ,...) Index -
       for every dimension that has length > 1, is not averaged across, and is not used as the X Axis, there is an
       slider to indicate which index along that dimension to use.
    """


    def initUI(self):
        # Widgets
        self.addWidget('NonExclusivePushButtons', 'M display',
                       buttons=['Mx','My','Mz','Mt','|M|'])
        self.addWidget('NonExclusivePushButtons', 'Average Across:',
                       buttons=['T1','T2','FQ','X0','Y0','Z0','Vx','Vy','Vz'], val=0)
        self.addWidget('ExclusivePushButtons', 'X Axis',
                       buttons=['T','T1','T2','FQ','X0','Y0','Z0','Vx','Vy','Vz'], val=0)

        self.addWidget('Slider', 'T  index')
        self.addWidget('Slider', 'T1 index')
        self.addWidget('Slider', 'T2 index')
        self.addWidget('Slider', 'FQ index')
        self.addWidget('Slider', 'X0 index')
        self.addWidget('Slider', 'Y0 index')
        self.addWidget('Slider', 'Z0 index')
        self.addWidget('Slider', 'Vx index')
        self.addWidget('Slider', 'Vy index')
        self.addWidget('Slider', 'Vz index')

        # IO Ports
        self.addInPort('M_in', 'NPYarray',obligation=gpi.REQUIRED)
        self.addOutPort('M_out', 'NPYarray')

        return 0

    def validate(self):
        '''This function runs before compute() as a GPI_APPLOOP exec-type.
        Here, widgets (bounds, limits, etc...) can be modified to ensure they
        are correctly validated before the widget values are used in the
        compute() routine -where widgets are buffered and any modifications are
        applied after compute() runs.
        '''

        m_in = self.getData('M_in')
        mdim = list(m_in.shape)
        maxt1 = mdim[2]
        maxt2 = mdim[3]
        maxfq = mdim[4]
        maxx0 = mdim[5]
        maxy0 = mdim[6]
        maxz0 = mdim[7]
        maxvx = mdim[8]
        maxvy = mdim[9]
        maxvz = mdim[10]

        avex = self.getVal('Average Across:')
        xaxis = self.getVal('X Axis')

        # Set slider indices for dimensions to be averaged to 1
        # Can't set x axis for a dimension to be averaged across
        #   so if this happens, set x axis to 0
        if avex:
          if (np.array(avex) == 0).any(): # T1
            maxt1 = 1
            if xaxis==1:
              xaxis = 0
              self.setAttr('X Axis',val = 0)
          if (np.array(avex) == 1).any(): # T2
            maxt2 = 1
            if xaxis==2:
              xaxis = 0
              self.setAttr('X Axis',val = 0)
          if (np.array(avex) == 2).any(): # FQ
            maxfq = 1
            if xaxis==3:
              xaxis = 0
              self.setAttr('X Axis',val = 0)
          if (np.array(avex) == 3).any(): # X0
            maxx0 = 1
            if xaxis==4:
              xaxis = 0
              self.setAttr('X Axis',val = 0)
          if (np.array(avex) == 4).any(): # Y0
            maxy0 = 1
            if xaxis==5:
              xaxis = 0
              self.setAttr('X Axis',val = 0)
          if (np.array(avex) == 5).any(): # Z0
            maxz0 = 1
            if xaxis==6:
              xaxis = 0
              self.setAttr('X Axis',val = 0)
          if (np.array(avex) == 6).any(): # Vx
            maxvx = 1
            if xaxis==7:
              xaxis = 0
              self.setAttr('X Axis',val = 0)
          if (np.array(avex) == 7).any(): # Vy
            maxvy = 1
            if xaxis==8:
              xaxis = 0
              self.setAttr('X Axis',val = 0)
          if (np.array(avex) == 8).any(): # Vz
            maxvz = 1
            if xaxis==9:
              xaxis = 0
              self.setAttr('X Axis',val = 0)

        self.setAttr('T  index',max=mdim[1]-2)
        self.setAttr('T1 index',max=maxt1-1)
        self.setAttr('T2 index',max=maxt2-1)
        self.setAttr('FQ index',max=maxfq-1)
        self.setAttr('X0 index',max=maxx0-1)
        self.setAttr('Y0 index',max=maxy0-1)
        self.setAttr('Z0 index',max=maxz0-1)
        self.setAttr('Vx index',max=maxvx-1)
        self.setAttr('Vy index',max=maxvy-1)
        self.setAttr('Vz index',max=maxvx-1)

        # Slider Visibility
        vist1 = vist2 = visfq = True
        visx0 = visy0 = visz0 = True
        visvx = visvy = visvz = True
        if maxt1 == 1 or xaxis == 1:
          vist1 = False
        if maxt2 == 1 or xaxis == 2:
          vist2 = False
        if maxfq == 1 or xaxis == 3:
          visfq = False
        if maxx0 == 1 or xaxis == 4:
          visx0 = False
        if maxy0 == 1 or xaxis == 5:
          visy0 = False
        if maxz0 == 1 or xaxis == 6:
          visz0 = False
        if maxvx == 1 or xaxis == 7:
          visvx = False
        if maxvy == 1 or xaxis == 8:
          visvy = False
        if maxvz == 1 or xaxis == 9:
          visvz = False

        self.setAttr('T  index',visible=xaxis!=0)
        self.setAttr('T1 index',visible=vist1)
        self.setAttr('T2 index',visible=vist2)
        self.setAttr('FQ index',visible=visfq)
        self.setAttr('X0 index',visible=visx0)
        self.setAttr('Y0 index',visible=visy0)
        self.setAttr('Z0 index',visible=visz0)
        self.setAttr('Vx index',visible=visvx)
        self.setAttr('Vy index',visible=visvy)
        self.setAttr('Vz index',visible=visvz)

        return 0

    def compute(self):
        '''This is where the main algorithm should be implemented.
        '''

        # New data
        m_in = self.getData('M_in')
        mdim = list(m_in.shape)

        xaxis = self.getVal('X Axis')

        idim = np.zeros(10, dtype=int)
        idim[1] = int(self.getVal('T1 index'))
        idim[2] = int(self.getVal('T2 index'))
        idim[3] = int(self.getVal('FQ index'))
        idim[4] = int(self.getVal('X0 index'))
        idim[5] = int(self.getVal('Y0 index'))
        idim[6] = int(self.getVal('Z0 index'))
        idim[7] = int(self.getVal('Vx index'))
        idim[8] = int(self.getVal('Vy index'))
        idim[9] = int(self.getVal('Vz index'))

# X AXIS
        if xaxis == 0: # Time
          xval = m_in[6,slice(1,mdim[1]),0,0,0,0,0,0,0,0,0]
        if xaxis == 1: # T1
          xval = m_in[1,0,:,0,0,0,0,0,0,0,0]
        if xaxis == 2: # T2
          xval = m_in[2,0,0,:,0,0,0,0,0,0,0]
        if xaxis == 3: # FQ
          xval = m_in[6,0,0,0,:,0,0,0,0,0,0]
        if xaxis == 4: # X0
          xval = m_in[3,1,0,0,0,:,0,0,0,0,0]
        if xaxis == 5: # Y0
          xval = m_in[4,1,0,0,0,0,:,0,0,0,0]
        if xaxis == 6: # Z0
          xval = m_in[5,1,0,0,0,0,0,:,0,0,0]
        if xaxis == 7: # Vx
          xval = m_in[3,0,0,0,0,0,0,0,:,0,0]
        if xaxis == 8: # Vy
          xval = m_in[4,0,0,0,0,0,0,0,0,:,0]
        if xaxis == 9: # Vz
          xval = m_in[5,0,0,0,0,0,0,0,0,0,:]

# Start with Y=0 X Axis to plot
        x0 = xval[:,np.newaxis]
        xdim = list(xval.shape)[0]
        y0 = np.zeros(xdim)[:,np.newaxis]

# Now add X=0 Y axis to plot
        x0 = np.append(x0,np.linspace(xval[0],xval[0],xdim)[:,np.newaxis],axis=1)
        y0 = np.append(y0,np.linspace(-1.,1.,xdim)[:,np.newaxis],axis=1)

# Y Axis for M
  # Average where Desired
        avex = self.getVal('Average Across:')
        m_av = m_in[:3,:,:,:,:,:,:,:,:,:,:]
        index_to_ave = 2
        for i in range(0,9):
          if (np.array(avex) == i).any(): # T1
            m_av = np.average(m_av,axis=index_to_ave)
          else:
            index_to_ave += 1

  # Set up outvals, from which Y axis will be picked
    # Always pass first dimension
        xi = [slice(None)]

    # Special case for 2nd (time) dimension
        if xaxis == 0:
          xi.append(slice(1,mdim[1]))
        else:
          xi.append(self.getVal('T  index')+1)

    # Remaining Dimensions
        for i in range(1,10):
          if xaxis == i:                      # Pass the x axis
            xi.append(slice(None))
          elif not (np.array(avex) == i-1).any(): # Slice this axis if it hasn't already been averaged
            xi.append(idim[i])

        outvals = m_av[xi]

  # No assign Y Axis
        mdisp = self.getVal('M display')
        if mdisp:
          if (np.array(mdisp) == 0).any(): # Mx
            x0 = np.append(x0,xval[:,np.newaxis],axis=1)
            y0 = np.append(y0,outvals[0,:,np.newaxis],axis=1)
          if (np.array(mdisp) == 1).any(): # My
            x0 = np.append(x0,xval[:,np.newaxis],axis=1)
            y0 = np.append(y0,outvals[1,:,np.newaxis],axis=1)
          if (np.array(mdisp) == 2).any(): # Mz
            x0 = np.append(x0,xval[:,np.newaxis],axis=1)
            y0 = np.append(y0,outvals[2,:,np.newaxis],axis=1)
          if (np.array(mdisp) == 3).any(): # Mt
            x0 = np.append(x0,xval[:,np.newaxis],axis=1)
            tval = np.sqrt(outvals[0,:]*outvals[0,:] + outvals[1,:]*outvals[1,:])
            y0 = np.append(y0,tval[:,np.newaxis],axis=1)
          if (np.array(mdisp) == 4).any(): # |M|
            x0 = np.append(x0,xval[:,np.newaxis],axis=1)
            tval = np.sqrt(outvals[0,:]*outvals[0,:] + outvals[1,:]*outvals[1,:] + outvals[2,:]*outvals[2,:])
            y0 = np.append(y0,tval[:,np.newaxis],axis=1)

        m_out = np.append(x0[:,:,np.newaxis],y0[:,:,np.newaxis],axis=2)

        self.setData('M_out', m_out)

        return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart, Jim Pipe
# Date: 2012oct28, 2013Dec


import numpy as np
import gpi

class ExternalNode(gpi.NodeAPI):
    """A visual for-loop that produces floats and integers for a predefined limits and steps.
    Module will output an number, wait for all other modules to run, then output the next integer
    Also allows generation of uniformly distributed random numbers, and manually-entered numbers
    A second "bound" mode is employed when input port is populated for spanning the bounds of a given dim.

    INPUTS:
    in_array: When populated, user specifies dimension, and can automatically or manually change index
              in the range of that dimension

    OUTPUTS:
    int_out - integer output, given as round(float_out)
    dict_for_reduce - passes the widget type necessary to hook IntegerLoop up to Reduce widget InPort
    float_out - floating point output

    WIDGETS:
    Dimension - when in_array populated, lets user pick dimension to for step to span
    Randomize - output value is either RANDOM or LINEAR, the latter = Minimum + step*(Step Size)
    LOOP - when selected, step runs from 0 to "Number of Steps"-1
    Continuous - when selected, step returns to 0 after reaching "Number of Steps" and continues to run
    Minimum - value corresponding to step=0 for LINEAR mode, or minimum range of output for RANDOM mode
    Maximum - value corresponding to last step for LINEAR mode, or maximum range of output for RANDOM mode
    Step Size - change in value per step for LINEAR mode
    Number of Steps - number of iterations between Minimum and Maximum
    step - current index
    Value - current value presented at output, either a random number between Minimum and Maximum for RANDOM mode
            or equal to Minimum + step*(Maximum-Minimum)/("Number of Steps" - 1)
    """

    def execType(self):
        return gpi.GPI_THREAD

    def initUI(self):

        # Widgets
        self.addWidget('Slider', 'Dimension', val=0, visible = False)
        self.addWidget('PushButton', 'Randomize', button_title='LINEAR', toggle=True)
        self.addWidget('PushButton', 'LOOP', button_title='OFF', val=0, toggle=True)
        self.addWidget('PushButton', 'Continuous', button_title='OFF', val=0, toggle=True)
        self.addWidget('DoubleSpinBox', 'Minimum', val=0, decimals=5)
        self.addWidget('DoubleSpinBox', 'Maximum', val=0, decimals=5)
        self.addWidget('DoubleSpinBox', 'Step Size', val=0, decimals=5)
        self.addWidget('SpinBox', 'Number of Steps', val=2, min=2)
        self.addWidget('Slider', 'step', val=0)
        self.addWidget('DoubleSpinBox', 'Value', val=0, decimals=5)

        # IO Ports
        self.addInPort('in_array', 'NPYarray',obligation=gpi.OPTIONAL)
        self.addOutPort('int_out', 'INT')
        self.addOutPort('dict_for_Reduce', 'DICT')
        self.addOutPort('float_out', 'FLOAT')

    def validate(self):

        # change button Titles with toggle
        status = self.getVal('LOOP')
        loop_over = self.getVal('Continuous')

        if status:
          self.setAttr('LOOP', button_title='ON')
        else:
          self.setAttr('LOOP', button_title='OFF')
        
        if loop_over:
          self.setAttr('Continuous', button_title='ON')
        else:
          self.setAttr('Continuous', button_title='OFF')

          # ONLY CHANGE THINGS IF STATUS IS OFF...
          indat = self.getData('in_array')
          if indat is None:
            # NO INPUT
            self.setAttr('Dimension', visible = False)
            self.setAttr('Randomize', visible = True)
            self.setAttr('Minimum', visible = True)
            self.setAttr('Maximum', visible = True)
            self.setAttr('Number of Steps', visible = True)
            self.setAttr('Value', visible = True)
            randomize = self.getVal('Randomize')
            if randomize:
              self.setAttr('Randomize', button_title='RANDOM')
              self.setAttr('Step Size', visible = False)
            else:
              self.setAttr('Randomize', button_title='LINEAR')
              self.setAttr('Step Size', visible = True)

            # Reconcile min,max,nsteps,stepsize
            minval = self.getVal('Minimum')
            maxval = self.getVal('Maximum')
            nsteps = self.getVal('Number of Steps')
            stepsize = self.getVal('Step Size')

            # make sure minval + nsteps*stepsize = maxval
            if 'Step Size' in self.widgetEvents():
              if stepsize != 0:
                nsteps = max(2,round(abs((maxval-minval)/stepsize)) + 1)
              else:
                nsteps = 2
              self.setAttr('Number of Steps',quietval=nsteps)
            self.setAttr('Step Size',quietval=(maxval-minval)/float(nsteps-1))
            self.setAttr('step',max = nsteps-1)
            self.setAttr('Value',min=minval,max=maxval)
          else:
            # INPUT, act in "Bounds" mode
            self.setAttr('Dimension', visible = True,min = -indat.ndim,max=indat.ndim-1)
            self.setAttr('Randomize', visible = False)
            self.setAttr('Minimum', visible = False)
            self.setAttr('Maximum', visible = False)
            self.setAttr('Step Size', visible = False)
            self.setAttr('Value', visible = False)
            self.setAttr('Number of Steps', visible = False)

            ndim = self.getVal('Dimension')
            self.setAttr('step',min=0,max=indat.shape[ndim]-1)

        return 0

    def compute(self):

        import numpy as np

        indat = self.getData('in_array')
        status = self.getVal('LOOP')
        cstep = self.getVal('step')
        loop_over = self.getVal('Continuous')

        # Mode depends on whether input port is populated
        if indat is None:
          minv = self.getVal('Minimum')
          maxv = self.getVal('Maximum')
          random = self.getVal('Randomize')
          nstep = self.getVal('Number of Steps')
          stepsize = (maxv-minv)/float(nstep-1)
        else:
          ndim = self.getVal('Dimension')
          minv = 0
          maxv = indat.shape[ndim]-1
          random = False
          nstep = indat.shape[ndim]
          stepsize = 1.

        if status:
          # if ON, Advance Step unless LOOP just pressed
          if 'LOOP' in self.widgetEvents():
            cstep = 0
            self.setAttr('step',quietval=cstep)
          else:
            if cstep < nstep-1:
              cstep += 1
              self.setAttr('step',quietval=cstep)
            else:
                if loop_over:
                    cstep = 0
                    self.setAttr('step', quietval=cstep)
                else:
                    self.setAttr('LOOP', quietval=False, button_title='OFF')
                    status = False
                    self.setAttr('Continuous', quietval=False, button_title='OFF')
                    loop_over = False

        # If user changed value manually, reflect that
        if 'Value' in self.widgetEvents():
          cvalue = self.getVal('Value')
          if stepsize != 0 and not random:
            cstep = round((cvalue-minv)/stepsize)
            cvalue = minv + cstep*stepsize
            self.setAttr('step',quietval=cstep)
        else:
        # Otherwise update values appropriately
          if random:
            cvalue = minv + (maxv-minv)*np.random.random()
          else:
            cvalue = minv + cstep*stepsize

        self.setAttr('Value',quietval=cvalue)
        self.setData('int_out', int(round(cvalue)))
        self.setData('float_out', cvalue)

        # set dict output to be used for Reduce module
        dict_reduce = {}
        dict_reduce['selection'] = 2
        dict_reduce['width'] = 1
        dict_reduce['ceiling'] = round(cvalue)
        dict_reduce['center'] = round(cvalue)
        dict_reduce['floor'] = round(cvalue)
        self.setData('dict_for_Reduce', dict_reduce)

        # logic for looping
        if status and (cstep <= nstep-1):
          self.setReQueue(True)
        else:
          self.setReQueue(False)
          self.setAttr('LOOP',quietval=False, button_title='OFF')

        return(0)



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Dallas Turley


import os, sys
import numpy as np
import gpi

class ExternalNode(gpi.NodeAPI):
    """Read images into GPI as numpy arrays. Many image file types are supported by PIL (Python Imaging Library)
    The file types specifically implemented in this node are PNG, JPEG, TIFF and BMP. PIL supports other image
    file types, but the node's ability to read these types into GPI are not guaranteed.
    For more information on supported file types, see
    http://www.pythonware.com/products/pil
    http://effbot.org/imagingbook/pil-index.htm

    OUTPUT: Numpy array read from image file.

    WIDGETS: I/O Info - Gives info on data file and type
    File Browser - button to launch file browser, and typein widget if the pathway is known
    Gray Scale - button to flatten the last dimension of jpg and png images.
        If toggle is on, output is a grey-scale float32 array with the same dimensions as the original image
        If toggle is off, output is a uint8 array with the last dimension BGRA
    """
    def execType(self):
        return gpi.GPI_PROCESS #this is safest

    def initUI(self):
        self.addWidget('TextBox', 'I/O Info:')
        self.addWidget('OpenFileBrowser', 'File Browser', button_title='Browse', caption='Open File', filter='png (*.png);;jpg (*.jpg);;tiff (*.tiff);;all (*)')
        self.addWidget('PushButton', 'Gray Scale', toggle=True)
        self.addWidget('PushButton', 'Swap Colors (RGB -> BGR)', toggle=True, val=0, button_title='RGB')

        self.addOutPort(title='out', type='NPYarray')

        self.URI = gpi.TranslateFileURI

    def validate(self) :
        if self.getVal('Swap Colors (RGB -> BGR)') :
            self.setAttr('Swap Colors (RGB -> BGR)', button_title="BGR")
        else :
            self.setAttr('Swap Colors (RGB -> BGR)', button_title="RGB")

        fname = self.URI(self.getVal('File Browser'))
        self.setDetailLabel(fname)

        return 0

    def compute(self):
        import numpy as np
        import time
        from PIL import Image

        flat = self.getVal('Gray Scale')
        swap = self.getVal('Swap Colors (RGB -> BGR)')

        #start file browser
        fname = self.URI(self.getVal('File Browser'))

        # ignore this case
        if fname == '':
            return 0

        #check that the path actually exists
        if not os.path.exists(fname):
            self.log.warn("Path does not exist: "+str(fname))
            return 0

        #show some file stats
        fstats = os.stat(fname)
        #creation
        ctime = time.strftime('%d/%m/20%y', time.localtime(fstats.st_ctime))
        # mod time
        mtime = time.strftime('%d/%m/20%y', time.localtime(fstats.st_mtime))
        # access time
        atime = time.strftime('%d/%m/20%y', time.localtime(fstats.st_atime))
        # filesize
        fsize = fstats.st_size
        # user id
        uid = fstats.st_uid
        # group id
        gid = fstats.st_gid

        # read the data
        img = Image.open(fname)
        if flat:
            img = img.convert('L')
        elif img.mode == 'RGB' :
            img = img.convert('RGBA')

        imgarr = np.array(img)

        if swap and imgarr.shape[-1] > 1 :
            red = imgarr[:,:,2]
            blu = imgarr[:,:,1]
            grn = imgarr[:,:,0]
            if imgarr.shape[-1] == 4 :
                alf = imgarr[:,:,3]
            else :
                alf = 255. * np.ones((blu.shape),dtype=np.uint8)

            out = np.zeros(imgarr.shape,dtype=imgarr.dtype)
            out[:,:,0] = red
            out[:,:,1] = blu
            out[:,:,2] = grn
            out[:,:,3] = alf
        else :
            out = imgarr

        d1 = list(out.shape)
        info = "created: "+str(ctime)+"\n" \
               "accessed: "+str(atime)+"\n" \
               "modified: "+str(mtime)+"\n" \
               "UID: "+str(uid)+"\n" \
               "GID: "+str(gid)+"\n" \
               "file size (bytes): "+str(fsize)+"\n" \
               "dimensions: "+str(d1)+"\n" \
               "type: "+str(out.dtype)+"\n" \
               "output color band: RGBA"+"\n"
        self.setAttr('I/O Info:', val=info)

        self.setData('out', out)

        return 0

